
#include <xs1.h>
#include <platform.h>
#include <debug_print.h>
#include <xscope.h>
#include <xclib.h>

#include "quadflash.h"
#include "quadflashlib.h"

#include "pll_timer.h"

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

on tile[0]: fl_QSPIPorts qspi_ports =
{
  XS1_PORT_1B,
  XS1_PORT_1C,
  XS1_PORT_4B,
  XS1_CLKBLK_1
};


on tile[0]: port p_err = XS1_PORT_1A;


#define FLASH_PAGE_SZ 256

static uint32_t g_flashSize = 0;
static uint32_t g_pageSize = 0;
static uint32_t g_sectorSize = 0;
static uint8_t test_buf[FLASH_PAGE_SZ+16];
static uint8_t g_scratch_buf[FLASH_PAGE_SZ+16];

static uint32_t n_pass = 0;
static uint32_t n_fail = 0;


int flash_write_page(uint32_t addr, unsigned char data[], int num_bytes)
{
    int r;
    if ((addr % g_sectorSize) == 0)
    {
        // If this is the first page of a sector, erase the sector.
        while(fl_getBusyStatus())
            {;}
        
        r = fl_eraseSector(addr / g_sectorSize);
        if ( r ) {
            debug_printf("erase sector failed (%d)\n", r);
            return r;
        }
    }
    
    while(fl_getBusyStatus())
        {;}
    
    r = fl_writePage(addr, data);
    if ( r )
    {
        debug_printf("write page failed (%d)\n", r);
        return r;
    }
    
    while(fl_getBusyStatus())
        {;}
    
    return 0;
}

int flash_write_and_verify(uint32_t addr, unsigned char data[], int num_bytes)
{
    int r = 0;
    int n = 3;
    do {
        r = flash_write_page(addr, data, num_bytes);
        if(r)
        {
            return 0;
        }
        
        r = fl_readPage(addr, g_scratch_buf);
        if (r)
        {
            return 0;
        }
        
        r = memcmp(data, g_scratch_buf, num_bytes);
        if(r != 0)
        {
            r = flash_write_page(addr, data, num_bytes);
            if(r)
            {
                return 0;
            }
        }
    } while(r != 0 && n--);
    return 0;
}

void init_flash_test(void)
{
    memset(test_buf, 0x5A, sizeof(test_buf));
    memset(g_scratch_buf, 0x5A, sizeof(g_scratch_buf));
    
    for(int k = 0; k < FLASH_PAGE_SZ; ++k)
    {
        test_buf[k] = k;
    }
    flash_write_and_verify(1024*1024, test_buf, FLASH_PAGE_SZ);
}


void test_task()
{
    timer tmr;
    unsigned t;
    
    xscope_register(0);
    xscope_config_io(XSCOPE_IO_BASIC);
    
    int r = fl_connect(qspi_ports);
    if (r) {
        debug_printf("couldn't connect to flash device\n");
        return;
    }
    
    g_flashSize =  fl_getFlashSize();
    g_sectorSize = fl_getSectorSize(0);
    debug_printf("flash is %d bytes (page = %d, sector = %d)\n", 
        g_flashSize, g_pageSize, g_sectorSize);
    
    init_flash_test();
    
    pll_ctrl_init();
    
    tmr :> t;
    t += REFCLK_HZ;
    
    while ( 1 ) {
        select {
            case tmr when timerafter(t) :> t:
            {
                memset(g_scratch_buf, 0x5A, sizeof(g_scratch_buf));
                fl_readPage(1024*1024, g_scratch_buf);
                if(memcmp(test_buf, g_scratch_buf, FLASH_PAGE_SZ) == 0)
                {
                    ++n_pass;
                    debug_printf("Page verify success, fail rate %d\n", (100*n_fail)/(n_fail + n_pass));
                }
                else
                {
                    p_err <: 1;
                    delay_microseconds(1);
                    p_err <: 0;
                    
                    ++n_fail;
                    debug_printf("Page verify failed, fail rate %d\n", (100*n_fail)/(n_fail + n_pass));
                    
                    uint8_t prev_byte = 0;
                    for(int j = 0; j < 16; ++j)
                    {
                        for(int k = 0; k < 16; ++k)
                        {
                            int crsr = j*16 + k;
                            uint8_t read_byte = g_scratch_buf[crsr];
                            // printf(" %02X:%02X:%d", test_buf[crsr], read_byte, read_byte - prev_byte);
                            // printf(" %02X:%d", read_byte, read_byte - prev_byte);
                            // printf(" %d", read_byte - prev_byte);
                            printf(" %02X%c", read_byte, ((read_byte - prev_byte) != 1)? '*' : ' ');
                            
                            prev_byte = read_byte;
                        }
                        printf("\n");
                    }
                    printf("\n\n");
                }
                
                uint32_t count;
                int32_t stamp;
                pll_ctrl_get_stats(count, stamp);
                printf("count=%lu stamp=%ld\n", count, stamp);
                
                t += REFCLK_HZ/10;
                
                break;
            }
        }
    }
}

int main()
{
    par {
        on tile[0]: test_task();
    }
    
    return 0;
}
